#!/usr/bin/env python
# -*- coding:utf-8 -*- 

import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F

import os
import sys
import time
import numpy as np

from tokens import *
from vocabulary import Vocabulary, createVocabularies
from util import loadData, getMinibatch, makeMinibatch, splitTrainDev, IdxLoader
from util_feature import NgramFeatures
import ner

from test import test
import accuracy


np.random.seed(0)
torch.manual_seed(0)


def train(
    model,
    optimizer : optim,
    word_vocab: Vocabulary,
    char_vocab: Vocabulary,
    tag_vocab : Vocabulary,
    args,
):

    ## get mini-batch size
    batch_size = args.batch_size
    device = torch.device(model._device)

    ## load training data 
    train_words, train_tags = loadData(args.train_path)

    ## if dev data is missing, use 20% of the training data as the dev
    if args.dev_path:
        dev_words, dev_tags = loadData(args.dev_path)
    else:
        print("create dev data from train data")
        train_words, train_tags, dev_words, dev_tags = splitTrainDev(train_words, train_tags)

    print("loaded train data: {} sentences".format(len(train_words)))
    print("loaded dev   data: {} sentences".format(len(dev_words)))

    ## the model stores the best f-measure on the dev
    model.f = -1
    n_worse = 0
    print(model)

    loader = IdxLoader(len(train_words))
    model.train()

    ## if an external dictionary is given, use it
    if args.dict_path and args.feature_n > 0:

        feat_handler = NgramFeatures(
            args.dict_path,
            args.feature_n,
        )

        print("use N-gram features")
        for n in range(1, args.feature_n+1):
            print("\t# of {}-grams: {: 5d}".format(n, len(feat_handler.ngrams[n])))
        print()

    else:
        feat_handler = None

    ## start training
    print("started training!")
    for iter_i in range(1, args.max_iter + 1):

        idx = loader.load(batch_size)

        ## get words and tags for mini-batch
        batch_words, batch_tags = getMinibatch(
            train_words,
            train_tags,
            idx,
        ) 

        ## prepare features
        if feat_handler:
            features = feat_handler.calcFeaturesBatch(batch_words)
            features = features.to(device)
        else:
            features = None

        ## create mini-batch
        batch_words, batch_word_lens, batch_word_mask, \
        batch_chars, batch_char_lens, batch_tags = makeMinibatch(
            batch_words,
            batch_tags,
            word_vocab,
            char_vocab,
            tag_vocab,
            device,
        )

        f_start = time.time()

        ## forward
        loss = model(
            batch_words,
            batch_chars,
            batch_tags,
            features,
            tag_vocab,
            batch_word_mask,
            batch_word_lens,
            batch_char_lens,
        )

        f_end = time.time()
        b_start = time.time()

        ## backward and update the parameters
        optimizer.zero_grad()
        loss.backward()
        if args.clipping is not None and args.clipping > 0:
            nn.utils.clip_grad_norm_(model.parameters(), args.clipping)
        optimizer.step()

        b_end = time.time()

        print("iter: {:>5d}, loss: {:10.4f}, forward: {: 2.2f}, backward: {: 2.2f}".format(
            iter_i,
            loss.tolist(),
            f_end - f_start,
            b_end - b_start,
        ))

        ## evaluate on the dev every iter_i iterations
        if iter_i % args.eval_iter == 0:

            model.eval()

            # calc accuracy on dev
            p, r, f = calcAccuracy(
                model, 
                word_vocab, 
                char_vocab,
                tag_vocab, 
                dev_words, 
                dev_tags,
                feat_handler,
            )

            model.train()

            ## save the parameters
            if model.f < f and args.model_path is not None:
                torch.save(
                    model.state_dict(),
                    args.model_path,
                )

            prev_f = model.f
            model.f = max(model.f, f)

            ## if the accuracy gets worse, halve the lr or finish the training
            if prev_f > f:

                n_worse += 1

                for param_group in optimizer.param_groups:
                    param_group["lr"] *= 0.5
                print("halved lr => {:.6f}".format(param_group["lr"]))

                if n_worse >= args.early_stopping:
                    print("finished training!")
                    print("the best F-measure on dev: {}".format(model.f))
                    sys.exit(0)


def calcAccuracy(
    model,
    word_vocab: Vocabulary, 
    char_vocab: Vocabulary, 
    tag_vocab : Vocabulary, 
    dev_words : list,
    dev_tags  : list,
    feat_handler,
):
    _device = model._device
    model._device = "cpu"
    model = model.to(torch.device(model._device))

    ## predict tags
    hyp_tags = test(
        model, 
        word_vocab, 
        char_vocab,
        tag_vocab, 
        dev_words, 
        feat_handler,
        to_stdout=False,
    )

    ref_list = [
        [ "{}/{}".format(w, t) for w, t in zip(words, tags) ]
        for words, tags in zip(dev_words, dev_tags)
    ]

    hyp_list = [
        [ "{}/{}".format(w, t) for w, t in zip(words, tags) ]
        for words, tags in zip(dev_words, hyp_tags)
    ]

    print("\nPrinting the first five inferences on the dev data")
    print("-" * 50)
    for h in hyp_list[:5]:
        print(" ".join(h))
    print()

    ## get precision, recall, and f-measure on dev
    p, r, f = accuracy.accuracy(ref_list, hyp_list, False)
    print()

    model._device = _device
    modle = model.to(torch.device(model._device))

    return p, r, f


if __name__ == "__main__":
    import os
    import pickle
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-g", "--gpu", type=int, default=-1, help="gpu id (use cpu if gpu<=-1)")

    parser.add_argument("--ne_list_path", type=str, default=None, help="NE list")
    parser.add_argument("--train_path", type=str, default=None, help="train data path")
    parser.add_argument("--dev_path", type=str, default=None, help="dev data path (if not, use 20 percent of the training data for development")
    parser.add_argument("--model_path", type=str, default=None, help="where to save model params")
    parser.add_argument("--vocab_path", type=str, default=None, help="where to save vocabs")

    parser.add_argument("--dict_path", type=str, default=None, help="external dictionary path (if not, train a model without external dictionary")
    parser.add_argument("--feature_n", type=int, default=0, help="the maximum n-gram feature size")

    parser.add_argument("-o", "--optimizer", type=str, default="Adam", help="optimizer type")
    parser.add_argument("--lr", type=float, default=1e-3, help="initial learning rate")
    parser.add_argument("--momentum", type=float, default=0.9, help="momentum (only for SGD)")
    parser.add_argument("--weight_decay", type=float, default=1e-6, help="weight decay ratio")
    parser.add_argument("--clipping", type=float, default=1.0, help="gradient clipping")

    parser.add_argument("-e", "--batch_size", type=int, default=10, help="epoch size")

    #parser.add_argument("--vocab_size", type=int, default=25000, help="vocab size")
    parser.add_argument("--freq_word", type=int, default=2, help="the minimum frequency for building word vocabulary")
    parser.add_argument("--freq_char", type=int, default=2, help="the minimum frequency for building char vocabulary")
    parser.add_argument("--hidden_word_size", type=int, default=256, help="word embedding size")
    parser.add_argument("--hidden_char_size", type=int, default=64, help="char embedding size")

    parser.add_argument("--max_iter", type=int, default=100000, help="the maximum number of training iters")
    parser.add_argument("--eval_iter", type=int, default=500, help="check accuracy on dev every eval_iter iters")
    parser.add_argument("--early_stopping", type=int, default=5, help="early stopping param")
    args = parser.parse_args()

    if args.gpu >= 0:
        torch.cuda.set_device(args.gpu)

    ## raise exception if a file does not exist
    for filename in (args.train_path, args.vocab_path):
        if filename is None:
            raise Exception("{} is None.".format(filename))

    ## create/load vocabularies
    if os.path.exists(args.vocab_path):
        word_vocab, char_vocab, tag_vocab = pickle.load(open(args.vocab_path, "rb"))
        print("loaded {}".format(args.vocab_path))
    else:
        word_vocab, char_vocab, tag_vocab = createVocabularies(
            file_path=args.train_path,
            ne_list_path=args.ne_list_path,
            freq_word=args.freq_word,
            freq_char=args.freq_char,
        )
        print("created {}".format(args.vocab_path))

        with open(args.vocab_path, "wb") as f:
            pickle.dump([word_vocab, char_vocab, tag_vocab], f)

    print("vocab (word) size: ", len(word_vocab.t2i))
    print("vocab (char) size: ", len(char_vocab.t2i))
    print("vocab (tag)  size: ", len(tag_vocab.t2i))
    print("\t" + " ".join(list(tag_vocab.t2i.keys())))

    ## initialize model parameters
    model = ner.BiLSTMCRF(
        len(word_vocab.t2i.keys()),
        len(char_vocab.t2i.keys()),
        args.hidden_word_size,
        args.hidden_char_size,
        len(tag_vocab.t2i.keys()),
        args.feature_n,
    )

    ## set a device where the model runs on
    model._device = "cuda:{}".format(args.gpu) if args.gpu >= 0 else "cpu"
    model = model.to(torch.device(model._device))

    ## prepare an optimizer
    if args.optimizer == "SGD":
        optimizer = optim.SGD(
            model.parameters(),
            lr=args.lr,
            weight_decay=args.weight_decay,
            momentum=args.momentum,
        )
    elif args.optimizer == "Adam":
        optimizer = optim.Adam(
            model.parameters(),
            lr=args.lr,
            weight_decay=args.weight_decay,
        )

    train(
        model,
        optimizer,
        word_vocab,
        char_vocab,
        tag_vocab,
        args,
    )


