#!/usr/bin/env python
# -*- coding:utf-8 -*-

import torch
from torch.autograd import Variable

import numpy as np

from tokens import *
from vocabulary import Vocabulary


## load raw data
def loadData(filename: str):
    words_list = []
    tags_list  = []

    for line in open(filename, "r").readlines():
        
        words = []
        tags  = []
        
        for chunk in line.rstrip().split(" "):
            chunk = chunk.split("/")
            assert len(chunk) == 2

            words += [chunk[0]]
            tags  += [chunk[1]]
            
        words_list += [words]
        tags_list  += [tags]

    return words_list, tags_list


## fill empty spaces in a mini-batch
def fill(
    batch: list,
    lens : list,
    pad  : int=-1,
):
    len_max = max(lens)

    return [
        batch[i] + [pad] * (len_max - lens[i])
        for i in range(len(lens))
    ]


## get a mask for a mini-batch
def getMask(lens: list):

    len_max = max(lens)
    
    return torch.ByteTensor(
        [
            [0] * lens[i] + [1] * (len_max - lens[i])
            for i in range(len(lens))
        ]
    )


## get a mini-batch from data
def getMinibatch(
    words,
    tags,
    idx,
):
    batch_words = [words[i] for i in idx]
    batch_tags  = [tags[i] for i in idx]

    sorted_pairs = [
        [s, t]
        for s, t in sorted(
            [ [_s, _t] for _s, _t in zip(batch_words, batch_tags) ],
            key=lambda x: len(x[0]), reverse=True,
        )
    ]

    batch_words = [x[0] for x in sorted_pairs]
    batch_tags  = [x[1] for x in sorted_pairs] 

    return batch_words, batch_tags


## prepare a mini-batch for calculation
def makeMinibatch(
    batch_words: list,
    batch_tags : list,
    word_vocab : Vocabulary,
    char_vocab : Vocabulary,
    tag_vocab  : Vocabulary,
    device     : torch.device,
):
    batch_word_lens = [len(x) for x in batch_words]

    max_len_batch = max(batch_word_lens)
    max_len_char = max([
        max([len(w) for w in words])
        for words in batch_words
    ])
    batch_size = len(batch_words)

    batch_chars = []
    batch_char_lens = []

    ## (batch_size x max(batch_word_lens), )
    for words in batch_words:
        words = words + [EOS] * (max_len_batch - len(words))
        for chars in words:
            batch_chars += [char_vocab.tokens2ids(chars)]
            batch_char_lens += [len(chars)]

    batch_chars = torch.LongTensor(
        fill(
            batch_chars,
            batch_char_lens,
            char_vocab.t2i[EOS],
        )
    ).to(device)

    batch_words = torch.LongTensor(
        fill(
            [word_vocab.tokens2ids(words) for words in batch_words],
            batch_word_lens,
            word_vocab.t2i[EOS],
            #word_vocab.t2i[PAD],
        )
    ).to(device)

    batch_tags = torch.LongTensor(
        fill(
            [tag_vocab.tokens2ids(tags) for tags in batch_tags],
            batch_word_lens,
            tag_vocab.t2i[STOP],
        )
    ).to(device)

    batch_word_mask = getMask(batch_word_lens).to(device)

    #return batch_words, batch_tags, batch_masks, batch_word_lens
    return batch_words, batch_word_lens, batch_word_mask, \
           batch_chars, batch_char_lens, batch_tags


## sort a list and returns the sorted one and an order
def getOrder(lens: list):

    lens = torch.LongTensor(lens)
    sorted_lens, order = torch.sort(lens, descending=True)
    return sorted_lens, order


## get an original order
def getOrderOrig(
    order: torch.LongTensor,
):
    pairs = [
        (o1, o2)
        for o1, o2 in zip(order.tolist(), list(range(len(order))))
    ]
    pairs.sort(key=lambda x: x[0])

    order_orig = [o2 for o1, o2 in pairs]
    return torch.LongTensor(order_orig)


## divide training data into training and development data
def splitTrainDev(
    words: list,
    tags : list,
    ratio: float=0.8, # how many are used for training
):
    size = len(words)

    indices = np.random.permutation(size)
    train_indices = indices[:int(size * ratio)]
    dev_indices = indices[int(size * ratio):]
    
    train_words = [words[i] for i in train_indices]
    train_tags = [tags[i] for i in train_indices]

    dev_words = [words[i] for i in dev_indices]
    dev_tags = [tags[i] for i in dev_indices]

    return train_words, train_tags, dev_words, dev_tags


## for getting indices of mini-batches
class IdxLoader():
    def __init__(self, size: int):
        self.size = size
        self.indexes = np.random.permutation(self.size).tolist()

    def getN(self, n):
        idx = self.indexes[:n]
        self.indexes = self.indexes[n:]
        
        if len(self.indexes) <= 0:
            self.indexes = np.random.permutation(self.size).tolist()
            
        return idx

    def load(self, x: int=128):
        idx = []

        while len(idx) != x:
            idx += self.getN(x - len(idx))

        return idx


